package com.cim.idm.framework.action.dowork;

import java.lang.reflect.Array;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.cim.idm.framework.IDMFrameServiceProxy;
import com.cim.idm.framework.data.EventInfo;
import com.cim.idm.framework.data.TransitionInfo;
import com.cim.idm.framework.exception.ErrorSignal;
import com.cim.idm.framework.exception.IDMFrameDBErrorSignal;
import com.cim.idm.framework.exception.ext.DuplicateNameSignal;
import com.cim.idm.framework.exception.ext.ExceptionKey;
import com.cim.idm.framework.exception.ext.ExceptionNotify;
import com.cim.idm.framework.exception.ext.FrameworkErrorSignal;
import com.cim.idm.framework.exception.ext.NotFoundSignal;
import com.cim.idm.framework.orm.MetaAttribute;
import com.cim.idm.framework.orm.OrmStopWatch;
import com.cim.idm.framework.orm.PersistentClass;
import com.cim.idm.framework.orm.SqlGenerator;
import com.cim.idm.framework.orm.info.DataInfo;
import com.cim.idm.framework.orm.info.KeyInfo;
import com.cim.idm.framework.orm.info.access.ObjectValueUserColumnsAccessor;
import com.cim.idm.framework.orm.info.access.UserColumnsAccessor;
import com.cim.idm.framework.orm.service.CommonService;
import com.cim.idm.framework.orm.support.OrmStandardEngineUtil;
import com.cim.idm.framework.util.object.ObjectUtil;
import com.cim.idm.framework.util.service.IDMFrameProperties;
import com.cim.idm.framework.util.time.TimeStampUtil;

/**
 * ActionProcessor will be called by ServiceDAO to process business <br>
 * ex: <br>
 *    LotService.create(key, eventinfo, createinfo); <br>
 *    will use ActionProcessor<LotKey,Lot,CreateInfo> 's dowork mehod to call CreateAction's create logic .<br>
 * Step: 
 * <ol>
 * <li> begin transaction <br></li>
 * <li> setEventinfo<br></li>
 * <li> setUserColumn<br></li>
 * <li> call ActionFramework to execute Action's Method<br></li>
 * <li> commit  or rollback if exception occur;<br></li>
 * </ol>
 * dowork function is not only one method, but three method as below:
 * <ul>
 * <li>dowork<br>
 *   Submit an SQL statement to the database and change a row of data</li>
 * <li>doworkForBatch<br>
 *   Submit a batch of SQL statements to the database and execute one by one in the database, each of which changes one row of data</li>
 * <li>doworkAtOnce<br>
 *   Submit an SQL statement to the database and change multiple rows of data at one time </li>
 * </ul>
 * 
 * 
 * @author 111079
 *
 * @param <KEY> KeyInfo
 * @param <DATA> DataInfo
 * @param <INFO> TransitionInfo
 */
public class ActionProcessor<KEY extends KeyInfo, DATA extends DataInfo, INFO extends TransitionInfo> {
	private static Log				log	= LogFactory.getLog(ActionProcessor.class);
	
	private CommonService<KEY,DATA> service;
	
	public ActionProcessor(CommonService<KEY, DATA> service) {
		super();
		this.service = service;
	}

	public CommonService<KEY, DATA> getService() {
		return service;
	}

	public void setService(CommonService<KEY, DATA> service) {
		this.service = service;
	}

	
	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @param eventInfo
	 *            com.cim.idm.framework.generic.info.EventInfo
	 * @param transitionInfo
	 *            extends com.cim.idm.framework.generic.info.TransitionInfo
	 * @param operationName
	 *            String
	 * @return DataInfo extends com.cim.idm.framework.orm.info.DataInfo
	 * @throws FrameworkErrorSignal
	 * @throws NotFoundSignal
	 * @throws DuplicateNameSignal
	 */
	@SuppressWarnings("unchecked")
	public DATA doWork(KEY keyInfo, EventInfo eventInfo, INFO transitionInfo, String operationName)
			throws FrameworkErrorSignal, NotFoundSignal, DuplicateNameSignal
	{
		DATA oldData;
		DATA newData;
		
		OrmStopWatch stopWatch = new OrmStopWatch(String.format("%s.%s DataKey=[%s]", this.getClass().getSimpleName(), operationName, ObjectUtil.getString(keyInfo)));
		stopWatch.start("B");
		IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
		stopWatch.stop();
		
		if (log.isInfoEnabled())
		{
			log.info("▶ START " + stopWatch.getId());
		}
		
		try
		{
			stopWatch.start("L");
			
			if ( operationName.startsWith("create") ) 
				oldData = (DATA) OrmStandardEngineUtil.createDataInfo(keyInfo);
			else 
				oldData = service.selectByKeyForUpdate(keyInfo);

			stopWatch.stop();
			//Set EnentInfo and UserColumns
			stopWatch.start("U");
			newData = (DATA) ObjectUtil.copyTo(oldData);

			ObjectUtil.setEventInfo(newData, eventInfo, transitionInfo);
			//transitionInfo extends UserColumnsAccessor , userColumn type only String.
			if(transitionInfo instanceof UserColumnsAccessor)
				ObjectUtil.setUserColumsToData(newData, ObjectUtil.getUserColumsValue(transitionInfo));
			//transitionInfo extends ObjectValueUserColumnsAccessor, userColumn type can be String, long , int , double , Date etc.
			else if(transitionInfo instanceof ObjectValueUserColumnsAccessor)
				ObjectUtil.setObjectUserColumnsToData(newData, ObjectUtil.getUserColumsObjectValue(transitionInfo));
			stopWatch.stop();
			//Execute Actions
			stopWatch.start("A");
			DATA result = (DATA)IDMFrameServiceProxy.getActionContainerManager().Execute( eventInfo.getBehaviorName(), newData.getClass().getSimpleName(), operationName, oldData, newData, eventInfo, transitionInfo);
			stopWatch.stop();

			boolean isRemoved = false;
			//Get Result
			stopWatch.start("S");
			if ( operationName.toLowerCase().startsWith("create") == false )
			{
				if ( operationName.toLowerCase().startsWith("remove") == false ) 
					service.update(newData);
				else isRemoved = true;
			}

			if ( isRemoved == false ) 
				result = service.selectByKey((KEY) result.getKey());
			stopWatch.stop();
			//Commit
			stopWatch.start("C");
			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
			stopWatch.stop();
			log.info("▶ ENDOK " + stopWatch.toString());
			
			return result;
		} 
		catch (Throwable e)
		{
			if (stopWatch.isRunning()) stopWatch.stop();
			//Rollback
			stopWatch.start("R");
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			if (stopWatch.isRunning()) stopWatch.stop();
			
			log.info("▶ ENDNG " + stopWatch.toString());
			
			throw ExceptionNotify.getNotifyException(e);
		}
	}
	
	@SuppressWarnings("unchecked")
	public DATA[] doWorkForBatch(KEY[] keyInfoList, EventInfo eventInfo, INFO[] transitionInfoList, String operationName)
			throws FrameworkErrorSignal, NotFoundSignal, DuplicateNameSignal
	{
		if(keyInfoList == null || keyInfoList.length == 0) {
			log.error(String.format("%s.%s DataKey=[%s]", this.getClass().getName(),operationName,ObjectUtil.getString(keyInfoList)));
			log.error("input value is null or empty");
			return null;
		}
		
		DATA[] oldDataList;
//		DATA[] newDataList = (DATA[]) new DataInfo[0];
		
		OrmStopWatch stopWatch = new OrmStopWatch(String.format("%s.%s DataKey=[%s]", this.getClass().getSimpleName(), operationName, ObjectUtil.getString(keyInfoList)));
		stopWatch.start("B");
		IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
		stopWatch.stop();
		
		if (log.isInfoEnabled())
		{
			log.info("▶ START " + stopWatch.getId());
		}
		
		try
		{
			stopWatch.start("L");
			
			if ( operationName.startsWith("create") ) 
				oldDataList = (DATA[]) OrmStandardEngineUtil.createDataInfoList(keyInfoList);
			else {
				///select for batch and get oldDataList in sequence of keyInfoList
				oldDataList = service.selectByKeyForUpdateForBatch(keyInfoList);
				List<KEY>  selectedKeyInfoList = new ArrayList<KEY>();
				for(DATA selectedData : oldDataList){
					selectedKeyInfoList.add((KEY)selectedData.getKey());
				}
				KEY[] keyInfoArray = (KEY[]) new KeyInfo[0];
				log.debug("selected result's keyinfolist sequence is " + ObjectUtil.getString(selectedKeyInfoList.toArray(keyInfoArray)));
				DATA[] sortedOldDataList = (DATA[]) OrmStandardEngineUtil.sortDataInfoListByKeyInfoList(oldDataList, keyInfoList);
				DATA[] oldDataListtemp = (DATA[]) Array.newInstance(oldDataList[0].getClass(), oldDataList.length);
				System.arraycopy(sortedOldDataList, 0, oldDataListtemp, 0, oldDataList.length);
				oldDataList = oldDataListtemp;
				selectedKeyInfoList = new ArrayList<KEY>();
				for(DATA selectedData : oldDataList){
					selectedKeyInfoList.add((KEY)selectedData.getKey());
				}
				log.debug("sorted result's keyinfolist sequence is " + ObjectUtil.getString(selectedKeyInfoList.toArray(keyInfoArray)));
			}

			stopWatch.stop();
			//Set EventInfo and UserColumns
			stopWatch.start("U");
			List<DATA> listFornewDataList = new ArrayList<DATA>();
			
			int nDataCount = 0;
			for(DATA object : oldDataList)
			{
				DATA data = (DATA) ObjectUtil.copyTo(object);
				ObjectUtil.setEventInfo(data, eventInfo, transitionInfoList[nDataCount]);
				//transitionInfo extends UserColumnsAccessor , userColumn type only String.
				if(transitionInfoList[nDataCount] instanceof UserColumnsAccessor)
					ObjectUtil.setUserColumsToData(data, ObjectUtil.getUserColumsValue(transitionInfoList[nDataCount]));
				//transitionInfo extends ObjectValueUserColumnsAccessor, userColumn type can be String, long , int , double , Date etc.
				else if(transitionInfoList[nDataCount] instanceof ObjectValueUserColumnsAccessor)
					ObjectUtil.setObjectUserColumnsToData(data, ObjectUtil.getUserColumsObjectValue(transitionInfoList[nDataCount]));
				
				listFornewDataList.add(data);
				nDataCount++;
			}
			
			
			DATA[] newDataList = (DATA[])listFornewDataList.toArray(oldDataList);

			//apply event info to newData
			
			stopWatch.stop();
			
			//Execute Actions
			stopWatch.start("A");
			DATA[] result = (DATA[])IDMFrameServiceProxy.getActionContainerManager().ExecuteForBatch(eventInfo.getBehaviorName(), newDataList[0].getClass().getSimpleName(), operationName, oldDataList, newDataList, eventInfo, transitionInfoList);
			stopWatch.stop();
			boolean isRemoved = false;
			//Get Result
			stopWatch.start("S");//Select Return Result
			if ( operationName.toLowerCase().startsWith("create") == false )
			{
				if ( operationName.toLowerCase().startsWith("remove") == false ) 
					service.update(newDataList);
				else isRemoved = true;
			}

			if ( isRemoved == false ) {
				for(int i = 0 ; i<result.length ; i++)
					keyInfoList[i] = (KEY) ObjectUtil.getKeyInfo(result[i]);
				result = service.selectByKeyForBatch(keyInfoList);
			}
			stopWatch.stop();
			//Commit
			stopWatch.start("C");
			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
			stopWatch.stop();
			log.info("▶ ENDOK " + stopWatch.toString());
			
			return result;
		} 
		catch (Throwable e)
		{
			if (stopWatch.isRunning()) stopWatch.stop();
			//Rollback
			stopWatch.start("R");
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			if (stopWatch.isRunning()) stopWatch.stop();
			
			log.info("▶ ENDNG " + stopWatch.toString());
			
			throw ExceptionNotify.getNotifyException(e);
		}
	}


	@SuppressWarnings("unchecked")
	public DATA[] doWorkAtOnce(KEY[] keyInfoList, EventInfo eventInfo, INFO transitionInfo, String operationName)
			throws FrameworkErrorSignal, NotFoundSignal, DuplicateNameSignal
	{
		if(keyInfoList == null || keyInfoList.length == 0) {
			log.error(String.format("%s.%s DataKey=[%s]", this.getClass().getName(),operationName,ObjectUtil.getString(keyInfoList)));
			log.error("input value is null or empty");
			return null;
		}
		
		DATA[] oldDataList = null;
		
		OrmStopWatch stopWatch = new OrmStopWatch(String.format("%s.%s DataKey=[%s]", this.getClass().getSimpleName(), operationName, ObjectUtil.getString(keyInfoList)));
		stopWatch.start("B"); //B -- Begin Transaction
		IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
		stopWatch.stop();
		
		if (log.isInfoEnabled())
		{
			log.info("▶ START " + stopWatch.getId());
		}
		try
		{
			if ( operationName.startsWith("create") ) //create / insert function should not use doWorkAtOnce
				throw new FrameworkErrorSignal(ExceptionKey.NotSupported_Function,operationName,"doWorkAtOnce");
			
			stopWatch.start("L");//L -- Lock
			
			int limitQty = 500;//default one sql process object's max quantity.
			//get object max process quantity from system property.
			String doWorkAtOnce_QtyLimit = System.getProperty( IDMFrameProperties.ONESQL_OBJ_QTY_LIMIT);
			if(StringUtils.isNotEmpty( doWorkAtOnce_QtyLimit ) && StringUtils.isNumeric( doWorkAtOnce_QtyLimit ))
				limitQty = Integer.parseInt(doWorkAtOnce_QtyLimit);
			int t = keyInfoList.length / limitQty;
			List<DATA[]> oldDataListArray = new ArrayList<DATA[]>();
			List<KEY[]> keyInfoListArray = new ArrayList<KEY[]>();
			DATA[] result = (DATA[])new DataInfo[keyInfoList.length];
			if(t==0){
				oldDataList = service.selectByKeyForUpdateForBatch(keyInfoList);
				oldDataListArray.add(oldDataList);
				keyInfoListArray.add(keyInfoList);
			}else{
				for(int i=0 ; i<t ; i++){
					KEY[] aKeyInfoArray = (KEY[])new KeyInfo[limitQty];
					System.arraycopy(keyInfoList, i*limitQty, aKeyInfoArray, 0, limitQty);
					oldDataList = service.selectByKeyForUpdateForBatch(aKeyInfoArray);
					oldDataListArray.add(oldDataList);
					keyInfoListArray.add(aKeyInfoArray);
				}
				if(keyInfoList.length-limitQty*t > 0){ //if keyInfoList.length-limitQty*t ==0 , should end 
					KEY[] aKeyInfoArray = (KEY[])new KeyInfo[keyInfoList.length-limitQty*t];
					System.arraycopy(keyInfoList, t*limitQty, aKeyInfoArray, 0, keyInfoList.length-limitQty*t);
					oldDataList = service.selectByKeyForUpdateForBatch(aKeyInfoArray);
					oldDataListArray.add(oldDataList);
					keyInfoListArray.add(aKeyInfoArray);
				}
			}
			
			stopWatch.stop();
			
			PersistentClass persistentClass = null;
			List<String> keyFieldNames = ObjectUtil.getFieldNames(keyInfoList[0]);
			
			int keyInfoListArraySize = keyInfoListArray.size();
			
			for(int i=0 ; i<keyInfoListArraySize ; i++){
				
				persistentClass = new PersistentClass();
				persistentClass.registerClass(oldDataList[0].getClass());
				for(KEY keyInfo : keyInfoListArray.get(i)){
					Map<String, Object > keyValueMap = ObjectUtil.getFieldValueMapOnlyKey(keyInfo);
					LinkedHashMap<String, Object> whereORConditionMap = new LinkedHashMap<String,Object>();
					for(String keyFieldName : keyFieldNames){
						whereORConditionMap.put(keyFieldName,keyValueMap.get(keyFieldName));
					}
					
					persistentClass.addWhereORConditionMap(whereORConditionMap);
				}
				//setEventInfo();
				ObjectUtil.setEventInfo(oldDataList[0], eventInfo, transitionInfo, persistentClass);
				//setUserColumn();
				//transitionInfo extends UserColumnsAccessor , userColumn type only String.
				if(transitionInfo instanceof UserColumnsAccessor)
					ObjectUtil.setUserColumsValue(oldDataList[0], ObjectUtil.getUserColumsValue(transitionInfo),persistentClass);
				//transitionInfo extends ObjectValueUserColumnsAccessor, userColumn type can be String, long , int , double , Date etc.
				else if(transitionInfo instanceof ObjectValueUserColumnsAccessor)
					ObjectUtil.setObjectUserColumnsValue(oldDataList[0], ObjectUtil.getUserColumsObjectValue(transitionInfo),persistentClass);
				
				stopWatch.start("A"); // A -- Action
				//do action
				IDMFrameServiceProxy.getActionContainerManager().ExecuteUpdateAtOnce(eventInfo.getBehaviorName(), oldDataList[0].getClass().getSimpleName(), operationName, oldDataListArray.get(i), eventInfo, transitionInfo,persistentClass);
//				doAction(action);
				stopWatch.stop();
				
				stopWatch.start("U"); // U -- Update
				
				boolean isRemoved = false;
				
				if ( operationName.toLowerCase().startsWith("create") == false )
				{
					if ( operationName.toLowerCase().startsWith("remove") == false ) {
						//update execute
						SqlGenerator sqlGenerator = new SqlGenerator(persistentClass);
						String updateSql = sqlGenerator.generateUpdateString();
						Object[] updateArgs = sqlGenerator.getUpdateArgs();
						IDMFrameServiceProxy.getSqlTemplate().update(updateSql, updateArgs);
						
						if(sqlGenerator.needHistory()){
							IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
							String insertSql = sqlGenerator.generateInsertSelectFromString();
							Object[] insertSelectArgs = sqlGenerator.getInsertSelectArgs();
							try {
								IDMFrameServiceProxy.getSqlTemplate().update(insertSql, insertSelectArgs);
								IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
							}catch(Throwable e) {
								IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
								if(e instanceof IDMFrameDBErrorSignal ) {
									IDMFrameDBErrorSignal ne = (IDMFrameDBErrorSignal) e;
									if (ErrorSignal.DuplicateNameSignal.equals(ne.getErrorCode())){
										boolean loopflag = false;
										do {
											try {
												duplicateNameExceptionWhenInsertHistoryByDoWorkAtOnce( eventInfo , persistentClass ,sqlGenerator, updateSql, insertSql, insertSelectArgs ) ;
											}catch(Throwable e1) {
												if(e instanceof IDMFrameDBErrorSignal ) {
													IDMFrameDBErrorSignal ne1 = (IDMFrameDBErrorSignal) e1;
													if (ErrorSignal.DuplicateNameSignal.equals(ne1.getErrorCode())){
														loopflag =true;
													}
												}else if ( e instanceof DuplicateNameSignal ) {
													loopflag =true;
												}else
													loopflag = false;
											}
										}while(loopflag);
										
									}
								}
								if ( e instanceof DuplicateNameSignal ) {
									boolean loopflag = false;
									do {
										try {
											duplicateNameExceptionWhenInsertHistoryByDoWorkAtOnce( eventInfo , persistentClass ,sqlGenerator, updateSql, insertSql, insertSelectArgs ) ;
										}catch(Throwable e1) {
											if(e instanceof IDMFrameDBErrorSignal ) {
												IDMFrameDBErrorSignal ne1 = (IDMFrameDBErrorSignal) e1;
												if (ErrorSignal.DuplicateNameSignal.equals(ne1.getErrorCode())){
													loopflag =true;
												}
											}else if ( e instanceof DuplicateNameSignal ) {
												loopflag =true;
											}else
												loopflag = false;
										}
									}while(loopflag);
								}
							}
						}
						
					}
					else isRemoved = true;
				}
				stopWatch.stop();
				//Get Result
				stopWatch.start("Z"); // S -- Select
				
				
				DATA[] aResult = null;
				if ( isRemoved == false ) 
					aResult = service.selectByKeyForBatch(keyInfoListArray.get(i));
					System.arraycopy(aResult, 0, result, i*limitQty, aResult.length);
				stopWatch.stop();
				
			}
			//Commit
			stopWatch.start("C"); // C -- Commit Transaction
			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
			stopWatch.stop();
			log.info("▶ ENDOK " + stopWatch.toString());
			return result;
			
		}catch (Throwable e)
		{
			if (stopWatch.isRunning()) stopWatch.stop();
			//Rollback
			stopWatch.start("R"); // Rollback Transaction
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			if (stopWatch.isRunning()) stopWatch.stop();
			
			log.info("▶ ENDNG " + stopWatch.toString());
			
			throw ExceptionNotify.getNotifyException(e);
		}
		
	}
	private void duplicateNameExceptionWhenInsertHistoryByDoWorkAtOnce(EventInfo eventInfo , PersistentClass persistentClass ,SqlGenerator sqlGenerator, String updateSql, String insertSql, Object[] insertSelectArgs ) {
		
		try {
			String newTimekey = "";
			String oldtimekey = (String) eventInfo.getEventTimeKey();
			String milimilisecondtimekey = oldtimekey.substring(oldtimekey.length()-6);
			
			int milimilisecondNum = Integer.parseInt(milimilisecondtimekey);
			milimilisecondNum = milimilisecondNum +1;
			milimilisecondtimekey = "" +milimilisecondNum;
			// when string size is less than six , fill in the front with 0
			for (int j = milimilisecondtimekey.length(); j < 6; j++) {
				milimilisecondtimekey = "0" + milimilisecondtimekey;
			}
			
			if(milimilisecondtimekey.length() >6) {
				milimilisecondtimekey = "000000";
				String realTime = oldtimekey.substring(0, 17);
				SimpleDateFormat sdfSimple = new SimpleDateFormat(TimeStampUtil.FORMAT_SIMPLE_DEFAULT);
				Date date = sdfSimple.parse(realTime);
				
				Calendar calendar = Calendar.getInstance();
				calendar.setTime(date);
				calendar.add(Calendar.SECOND, 1);
				Date newDate = calendar.getTime();
				
				String newTimekeySeconds = sdfSimple.format(newDate);
				newTimekey = newTimekeySeconds + milimilisecondtimekey;
			}else {
			
				newTimekey = oldtimekey.substring(0, oldtimekey.length()-6) + milimilisecondtimekey;
			}
	
			persistentClass.addUpdateMetaAttribute(new MetaAttribute("lastEventTimeKey", newTimekey));
			
			Object[]  updateArgs = sqlGenerator.getUpdateArgs();
			IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
			IDMFrameServiceProxy.getSqlTemplate().update(updateSql, updateArgs);
			IDMFrameServiceProxy.getSqlTemplate().update(insertSql, insertSelectArgs);
			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
		}catch(Throwable e) {
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			throw ExceptionNotify.getNotifyException(e);
		}
		
	}
}
