package com.cim.idm.framework.orm;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.dao.DataAccessException;
import org.springframework.util.CollectionUtils;

import com.cim.idm.framework.IDMFrameServiceProxy;
import com.cim.idm.framework.exception.ErrorSignal;
import com.cim.idm.framework.exception.IDMFrameDBErrorSignal;
import com.cim.idm.framework.exception.ext.NotFoundSignal;
import com.cim.idm.framework.orm.info.DataInfo;
import com.cim.idm.framework.orm.info.KeyInfo;
import com.cim.idm.framework.orm.support.OrmStandardEngineUtil;
import com.cim.idm.framework.util.object.ObjectUtil;

public class OrmStandardEngine<KEY extends KeyInfo, DATA extends DataInfo>
{
	private static Log				log	= LogFactory.getLog(OrmStandardEngine.class);

	public OrmStandardEngine()
	{
	}
	
	public List<DATA> transform(List resultList, Class clazz)
	{
		if (resultList.size() == 0)
			return null;

		Object result = OrmStandardEngineUtil.ormExecute(OrmStandardEngineUtil.createDataInfo(clazz), resultList);

		if (result instanceof List)
			return (List<DATA>) result;
		else
		{
			List<DATA> resultSet = new ArrayList<DATA>();
			resultSet.add((DATA) result);
			return resultSet;
		}
	}

	protected Object validateResultSet(List<Map<String,Object>> resultList, String keyData, String returnType) throws IDMFrameDBErrorSignal
	{
		Object result = OrmStandardEngineUtil.ormExecute(OrmStandardEngineUtil.createDataInfo(getClass()), resultList);
		if (result instanceof List)
		{
			return (List<DATA>) result;
		}
		else if (result instanceof DataInfo)
		{
			if (returnType.indexOf("list") >= 0)
			{
				List<DATA> resultSet = new ArrayList<DATA>();
				resultSet.add((DATA) result);
				return resultSet;
			}
			else
			{
				return result;
			}
		}
		else
		{
			throw new IDMFrameDBErrorSignal(ErrorSignal.CouldNotMatchData, keyData.toString());
		}
	}	
	
	public List<DATA> select(String condition, Object[] bindSet, Class<?> clazz) throws IDMFrameDBErrorSignal
	{
		String tableName = OrmStandardEngineUtil.getTableNameByClassName(clazz);
		if (tableName == null || tableName.length() == 0)
			tableName = OrmStandardEngineUtil.getTableNameByClassName(this.getClass());
		
		String sql = "select * from " + tableName;
		
		if (StringUtils.isNotEmpty(condition))
		{
			sql = OrmStandardEngineUtil.getConditionSql(sql, condition);
		}

		List<Map<String,Object>> resultList = queryForList(sql, bindSet);
		if (resultList.size() == 0)
			//throw new IDMFrameDBErrorSignal(ErrorSignal.NotFoundSignal, ObjectUtil.getString(bindSet),	SQLLogUtil.getLogFormatSqlStatement(sql, bindSet, log));
			throw new NotFoundSignal(getClass(), "select",  ObjectUtil.getString(bindSet),	SQLLogUtil.getLogFormatSqlStatement(sql, bindSet, log));

		Object result = OrmStandardEngineUtil.ormExecute(OrmStandardEngineUtil.createDataInfo(clazz), resultList);

		if (result instanceof List)
			return (List<DATA>) result;
		else if (result instanceof DataInfo)
		{
			List<DATA> resultSet = new ArrayList<DATA>();
			resultSet.add((DATA) result);
			return resultSet;
		}
		else
			throw new IDMFrameDBErrorSignal(ErrorSignal.CouldNotMatchData, ObjectUtil.getString(bindSet), SQLLogUtil.getLogFormatSqlStatement(sql, bindSet, log), condition);
	}

	public DATA selectByKeyForUpdate(KEY keyInfo) throws IDMFrameDBErrorSignal
	{
		return (DATA) execute("select.for.update", keyInfo);
	}
	
	public DATA[] selectByKeyForUpdateForBatch(KEY[] keyInfoList) throws IDMFrameDBErrorSignal
	{
		return (DATA[]) executeSelectBatch("select.for.update" , keyInfoList);
	}
	
	public DATA[] selectByKeyForUpdateForBatchWithOrder(KEY[] keyInfoList) throws IDMFrameDBErrorSignal
	{
		return (DATA[]) executeSelectBatchWithOrder("select.for.update",keyInfoList);
	}
	
	public DATA selectByKey(KEY keyInfo) throws IDMFrameDBErrorSignal
	{
		return (DATA) execute("select", keyInfo);
	}
	
	public List<DATA> selectByKey(List<KEY> keyInfoList) throws IDMFrameDBErrorSignal
	{
		DATA[] dataList =  (DATA[]) executeSelectBatch("select", keyInfoList.toArray());
		return Arrays.asList(dataList);
	}
	
	public DATA[] selectByKeyForBatch(KEY[] keyInfoList) throws IDMFrameDBErrorSignal
	{
		return (DATA[]) executeSelectBatch("select", keyInfoList);
	}
	
	public int update(DATA dataInfo, String condition, Object[] bindSet) throws IDMFrameDBErrorSignal
	{
		List<Object> objectValue = new ArrayList<Object>();
		String sql = OrmStandardEngineUtil.generateSqlStatement("update", dataInfo, objectValue);
		int idx = sql.indexOf(" where ");
		sql = sql.substring(0, idx + 1);
		sql = OrmStandardEngineUtil.getConditionSql(sql, condition);
		
		Map<String, Object> keyMap = ObjectUtil.getFieldValueMapOnlyKey(dataInfo);

		int argslength = objectValue.size()-keyMap.size()+bindSet.length;
		Object[] values = new Object[argslength];
		
		for(int i = 0 ; i < argslength; i++){
			if(i < objectValue.size()-keyMap.size()){
				values[i] = objectValue.get(i);
			}else
				values[i] = bindSet[i-objectValue.size()+keyMap.size()];
		}
		return executeQuery(sql, "update", values, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataInfo)));
	}

	public int update(DATA dataInfo) throws IDMFrameDBErrorSignal
	{
		return (Integer) execute("update", dataInfo);
	}

	public int[] update( List<DATA> dataInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "update", dataInfoList );
	}
	
	public int[] update( DATA[] dataInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "update", dataInfoList );
	}
		
	public int insert(DATA dataInfo) throws IDMFrameDBErrorSignal
	{
		return (Integer) execute("insert", dataInfo);
	}

	public int[] insert( List<DATA> dataInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "insert", dataInfoList );
	}
	
	public int[] insert( DATA[] dataInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "insert", dataInfoList );
	}
	
	public int delete(KEY keyInfo) throws IDMFrameDBErrorSignal
	{
		return (Integer) execute("delete", keyInfo);
	}

	public int delete(String condition, Object[] bindSet, Class clazz) throws IDMFrameDBErrorSignal
	{
		String tableName = OrmStandardEngineUtil.getTableNameByClassName(clazz);
		
		if (tableName == null || tableName.length() == 0)
			tableName = OrmStandardEngineUtil.getTableNameByClassName(this.getClass());

		String sql = "delete " + tableName;
		sql = OrmStandardEngineUtil.getConditionSql(sql, condition);
		
		return executeQuery(sql, "delete", bindSet, ObjectUtil.getString(condition, bindSet));
	}
	
	public int[] delete( List<KEY> keyInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "delete", keyInfoList );
	}
	
	public int[] delete( KEY[] keyInfoList ) throws IDMFrameDBErrorSignal
	{
		return executeBatch( "delete", keyInfoList );
	}

	protected int[] executeBatch( String queryType, List<?> dataObjectList ) throws IDMFrameDBErrorSignal
	{
		if ( CollectionUtils.isEmpty( dataObjectList ))
		{
			log.warn( "There is no Object to batch update." );
			return null;
		}
		
		List<Object> objectValue = new ArrayList<Object>();
		
		String sql = OrmStandardEngineUtil.generateSqlStatement(queryType, dataObjectList.get(0), objectValue);
		int[] resultSizes;
		
		
		List<Object[]> bindSetList = new ArrayList<Object[]>();
		for ( Object dataInfo : dataObjectList )
			bindSetList.add( OrmStandardEngineUtil.getBindSetObject(queryType, dataInfo ).toArray());
				
		
		if (queryType.equalsIgnoreCase("delete"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("update"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("insert"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else
			throw new IDMFrameDBErrorSignal(ErrorSignal.InvalidQueryType, queryType);

		return resultSizes;
	}
	
	protected int[] executeBatch( String queryType, KEY[] dataObjectList ) throws IDMFrameDBErrorSignal
	{
		if ( dataObjectList == null || dataObjectList.length ==0)
		{
			log.warn( "There is no Object to batch update." );
			return null;
		}
		
		List<Object> objectValue = new ArrayList<Object>();
		
		String sql = OrmStandardEngineUtil.generateSqlStatement(queryType, dataObjectList[0], objectValue);
		int[] resultSizes;
		
		
		List<Object[]> bindSetList = new ArrayList<Object[]>();
		for ( Object dataInfo : dataObjectList )
			bindSetList.add( OrmStandardEngineUtil.getBindSetObject(queryType, dataInfo ).toArray());
				
		
		if (queryType.equalsIgnoreCase("delete"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("update"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("insert"))
			resultSizes = IDMFrameServiceProxy.getSqlTemplate().updateBatch( sql, bindSetList);
		else
			throw new IDMFrameDBErrorSignal(ErrorSignal.InvalidQueryType, queryType);

		return resultSizes;
	}
	
	protected int[] executeBatch( String queryType, DATA[] dataObjectList ) throws IDMFrameDBErrorSignal
	{
		if ( dataObjectList == null || dataObjectList.length ==0)
		{
			log.warn( "There is no Object to batch update." );
			return null;
		}
		
		List<Object> objectValue = new ArrayList<Object>();
		
		String sql = OrmStandardEngineUtil.generateSqlStatement(queryType, dataObjectList[0], objectValue);
		int[] resultSizes;
		
		
		List<Object[]> bindSetList = new ArrayList<Object[]>();
		for ( Object dataInfo : dataObjectList )
			bindSetList.add( OrmStandardEngineUtil.getBindSetObject(queryType, dataInfo ).toArray());
				
		SqlTemplate sqlTemplate = IDMFrameServiceProxy.getSqlTemplate();
		if (queryType.equalsIgnoreCase("delete"))
			resultSizes = sqlTemplate.updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("update"))
			resultSizes = sqlTemplate.updateBatch( sql, bindSetList);
		else if (queryType.equalsIgnoreCase("insert"))
			resultSizes = sqlTemplate.updateBatch( sql, bindSetList);
		else
			throw new IDMFrameDBErrorSignal(ErrorSignal.InvalidQueryType, queryType);

		return resultSizes;
	}
	
	/*
	 * [Insert/Update] 
	 * 1. In the case of Field value of data is null
	 * 1.1. In the case of type of Date and TimeStamp, value is SYSDATE 
	 * 1.2. In the case of data type except 1.1, excluded from the target of the DML 
	 *   but When StandardType generated because String, Primitive numeric type is initialized, does not exist of null only if user does not set null randomly
	 * 1.3. In the case of Insert, have the the corresponding values if the Default value exist that defined in ObjectAttributeDef
	 * 
	 * 2. If the String type and UDF's value is "", erased the value of the corresponding field (finally Field of Database is null)
	 */
	protected Object execute(String queryType, Object dataObject) throws IDMFrameDBErrorSignal
	{
		List<Map<String,Object>> resultList = null;
		int resultSize = 0;
		String sql = "";
		List<Object> objectValue = new ArrayList<Object>();

		try
		{
			if (queryType.equalsIgnoreCase("select") || queryType.equalsIgnoreCase("select.for.update"))
			{
				KeyInfo keyInfo = OrmStandardEngineUtil.getKeyInfo(dataObject);
				if (ObjectUtil.isNullOrNullString(keyInfo))
					throw new IDMFrameDBErrorSignal(ErrorSignal.NullPointKeySignal, ObjectUtil.getString(keyInfo),	queryType, "Key can't be null");

				sql = OrmStandardEngineUtil.generateSqlStatement( queryType, dataObject, objectValue);
				
				resultList = queryForList(sql, objectValue.toArray());

				if (resultList.size() == 0)
					//throw new IDMFrameDBErrorSignal(ErrorSignal.NotFoundSignal, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, ObjectUtil.getFieldValueMapOnlyKey(dataObject), log));
					throw new NotFoundSignal(getClass(), queryType,  ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, objectValue.toArray(), log));
			}
			else
			{
				sql = OrmStandardEngineUtil.generateSqlStatement(queryType, dataObject, objectValue);
				
				if (queryType.equalsIgnoreCase("delete"))
					resultSize = update(sql, objectValue.toArray(), ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)));
				else if (queryType.equalsIgnoreCase("update"))
					resultSize = update(sql, objectValue.toArray(), ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)));
				else if (queryType.equalsIgnoreCase("insert"))
					resultSize = update(sql, objectValue.toArray(), ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)));
				else
					throw new IDMFrameDBErrorSignal(ErrorSignal.InvalidQueryType, queryType);

				if (resultSize == 0)
				{
					if (queryType.equalsIgnoreCase("delete") || queryType.equalsIgnoreCase("update"))
						//throw new IDMFrameDBErrorSignal(ErrorSignal.NotFoundSignal, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, ObjectUtil.getFieldValueMapOnlyKey(dataObject), log));
						throw new NotFoundSignal(getClass(), queryType,  ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, OrmStandardEngineUtil.getSelectOrDeleteBindObjects(dataObject), log));
					else
						throw new IDMFrameDBErrorSignal(ErrorSignal.InvalidQueryState, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), 
								SQLLogUtil.getLogFormatSqlStatement(sql, ObjectUtil.getFieldValueMapOnlyKey(dataObject), log));
				}
				
				return resultSize;
			}
		} 
		catch (DataAccessException e)
		{
			throw ErrorSignal.getNotifyException(e, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), sql);
		}
		
		return OrmStandardEngineUtil.ormExecute(dataObject, resultList);
	}
	
	protected Object[] executeSelectBatch(String queryType, Object[] dataObjectList) throws IDMFrameDBErrorSignal
	{
		List resultList = null;
		String sql = "";
		List<Object> objectValue = new ArrayList<Object>();
		
		//String queryType = "select.for.update";
		try
		{
			KeyInfo keyInfo = OrmStandardEngineUtil.getKeyInfo(dataObjectList[0]);
			if (ObjectUtil.isNullOrNullString(keyInfo))
				throw new IDMFrameDBErrorSignal(ErrorSignal.NullPointKeySignal, ObjectUtil.getString(keyInfo),	queryType, "Key can't be null");

			sql = OrmStandardEngineUtil.generateSqlStatementForBatch(queryType, dataObjectList, objectValue );
				
			resultList = queryForList(sql, objectValue.toArray());

			if (resultList.size() == 0)
				//throw new IDMFrameDBErrorSignal(ErrorSignal.NotFoundSignal, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, ObjectUtil.getFieldValueMapOnlyKey(dataObject), log));
				throw new NotFoundSignal(getClass(), queryType,  ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObjectList[0])), SQLLogUtil.getLogFormatSqlStatement(sql, OrmStandardEngineUtil.getSelectOrDeleteBindObjects(dataObjectList[0]), log));
			
			
		} 
		catch (DataAccessException e)
		{
			throw ErrorSignal.getNotifyException(e, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObjectList[0])), sql);
		}
		
		//DataInfo[] returnData = new DataInfo[0];
		//returnData = resultList.toArray()
		DataInfo[] result = OrmStandardEngineUtil.ormExecuteForBatch(dataObjectList, resultList);
		DataInfo dataObject = OrmStandardEngineUtil.createDataInfo(dataObjectList[0]);
		DATA[] resultData = (DATA[]) Array.newInstance(dataObject.getClass(), resultList.size());
		System.arraycopy(result, 0, resultData, 0, result.length);
		return resultData;
		
	}
	
	
	protected Object[] executeSelectBatchWithOrder(String queryType, Object[] dataObjectList) throws IDMFrameDBErrorSignal{
		List resultList = new ArrayList();
		
		String sql = "";
		for(Object dataObject :dataObjectList){
			List<Object> objectValue = new ArrayList<Object>();
			List resultListTemp ;
			KeyInfo keyInfo = OrmStandardEngineUtil.getKeyInfo(dataObject);
			if (ObjectUtil.isNullOrNullString(keyInfo))
				throw new IDMFrameDBErrorSignal(ErrorSignal.NullPointKeySignal, ObjectUtil.getString(keyInfo),	queryType, "Key can't be null");

			sql = OrmStandardEngineUtil.generateSqlStatement(queryType, dataObject, objectValue );
				
			String logFormatSql = SQLLogUtil.getLogFormatSqlStatement(sql, objectValue, log);
			try
			{
				SQLLogUtil.logBeforeExecutingQuery(logFormatSql, log);
				resultListTemp = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().queryForList(sql, objectValue.toArray());
			}catch (DataAccessException e)
			{
				throw ErrorSignal.getNotifyException(e, ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObjectList[0])), sql);
			}
			if (resultListTemp.size() == 0)
				throw new NotFoundSignal(getClass(), queryType,  ObjectUtil.getString(OrmStandardEngineUtil.getKeyInfo(dataObject)), SQLLogUtil.getLogFormatSqlStatement(sql, objectValue.toArray(), log));

			resultList.addAll(resultListTemp);	
		}
		SQLLogUtil.logAfterExecutingQuery(resultList, log);
		
		DataInfo[] result = OrmStandardEngineUtil.ormExecuteForBatch(dataObjectList, resultList);
		DataInfo dataObject = OrmStandardEngineUtil.createDataInfo(dataObjectList[0]);
		DATA[] resultData = (DATA[]) Array.newInstance(dataObject.getClass(), resultList.size());
		System.arraycopy(result, 0, resultData, 0, result.length);
		return resultData;
	}
	
	private int executeQuery(String sql, String operation,  Object[] bindSet, String keyString) throws IDMFrameDBErrorSignal
	{
		int resultSize = update(sql, bindSet, keyString);
		
		if (resultSize == 0)
			//throw new IDMFrameDBErrorSignal(ErrorSignal.NotFoundSignal, ObjectUtil.getString(bindSet), SQLLogUtil.getLogFormatSqlStatement(sql, bindSet, log));
			throw new NotFoundSignal(getClass(), operation,  ObjectUtil.getString(bindSet),	SQLLogUtil.getLogFormatSqlStatement(sql, bindSet, log));
		
		return resultSize;
	}

	private List<Map<String,Object>> queryForList(String sql, Object... args) throws IDMFrameDBErrorSignal
	{
		List<Map<String,Object>> resultList = null;
		String logFormatSql = SQLLogUtil.getLogFormatSqlStatement(sql, args, log);

		try
		{
			SQLLogUtil.logBeforeExecutingQuery(logFormatSql, log);

			if (ArrayUtils.isEmpty(args))
				resultList = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().queryForList(sql);
			else
			{
				if (args.length == 1 && args[0] == null)
					resultList = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().queryForList(sql);
				else
					resultList = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().queryForList(sql, args);
			}
		} 
		catch (DataAccessException e)
		{
			throw ErrorSignal.getNotifyException(e, ObjectUtil.getString(args), logFormatSql);
		}

		SQLLogUtil.logAfterExecutingQuery(resultList, log);
		return resultList ==null? new ArrayList<Map<String,Object>>():resultList;
	}

	private int update(String sql, Object[] args, String keyString) throws IDMFrameDBErrorSignal
	{
		String logFormatSql = SQLLogUtil.getLogFormatSqlStatement(sql, args, log);
		int result = 0;

		try
		{
			SQLLogUtil.logBeforeExecutingQuery(logFormatSql, log);

			if (args == null)
				result = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().update(sql);
			else
				result = IDMFrameServiceProxy.getSqlTemplate().getJdbcTemplate().update(sql, args);
		} 
		catch (DataAccessException e)
		{
			if (keyString != null)
				throw ErrorSignal.getNotifyException(e, keyString, logFormatSql);
			else
				throw ErrorSignal.getNotifyException(e, ObjectUtil.getString(args), logFormatSql);
		}

		SQLLogUtil.logAfterExecutingUpdate(result, log);
		return result;
	}
	
	
}
