package com.cim.idm.framework.orm.service;

import java.sql.Timestamp;
import java.util.List;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;

import com.cim.idm.framework.IDMFrameServiceProxy;
import com.cim.idm.framework.exception.ext.DuplicateNameSignal;
import com.cim.idm.framework.exception.ext.ExceptionNotify;
import com.cim.idm.framework.exception.ext.FrameworkErrorSignal;
import com.cim.idm.framework.exception.ext.InvalidStateTransitionSignal;
import com.cim.idm.framework.exception.ext.NotFoundSignal;
import com.cim.idm.framework.orm.info.DataInfo;
import com.cim.idm.framework.orm.info.HistoryDataInfo;
import com.cim.idm.framework.orm.info.HistoryKeyInfo;
import com.cim.idm.framework.orm.info.KeyInfo;
import com.cim.idm.framework.orm.support.OrmStandardEngineUtil;
import com.cim.idm.framework.util.object.ObjectUtil;
import com.cim.idm.framework.util.time.TimeUtils;

/**
 ****************************************************************************
 *  PACKAGE : com.cim.idm.framework.orm.service
 *  NAME    : CommonHistoryServiceDAO.java
 *  TYPE    : JAVA
 *  DESCRIPTION : Common Service for History-related Database Handling
 *
 ****************************************************************************
 */

public class CommonHistoryServiceDAO<HKEY extends HistoryKeyInfo, HDATA extends HistoryDataInfo> extends CommonServiceDAO<HKEY, HDATA>
		implements CommonHistoryService<HKEY, HDATA>
{
	public CommonHistoryServiceDAO()
	{

	}

	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @return DATA extends com.cim.idm.framework.orm.info.DataInfo
	 * @throws NotFoundSignal
	 * @throws InvalidStateTransitionSignal
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	@SuppressWarnings("unchecked")
	public HDATA selectLastOne(HKEY keyInfo)
			throws NotFoundSignal, InvalidStateTransitionSignal, DuplicateNameSignal, FrameworkErrorSignal
	{
		try
		{
			String sql = OrmStandardEngineUtil.generateSqlStatement("select", keyInfo);
			sql = OrmStandardEngineUtil.generateExtendedSqlStatement("select", keyInfo, "selectLastOne", sql);			
			List resultList = IDMFrameServiceProxy.getSqlTemplate().queryForList(sql, OrmStandardEngineUtil.getPrimaryNoTimeKeyForHistory(keyInfo,2));
			if (CollectionUtils.isEmpty(resultList))
			{
				throw new NotFoundSignal(getClass(), "selectLastOne", ObjectUtil.getString(keyInfo), sql);
			}

			HDATA result = (HDATA) validateResultSet(resultList, ObjectUtil.getString(keyInfo), "");

			return result;
		} catch (Throwable e)
		{
			throw ExceptionNotify.getNotifyException(e);
		}
	}

	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @param eventName
	 *            String
	 * @return DATA extends com.cim.idm.framework.orm.info.DataInfo
	 * @throws NotFoundSignal
	 * @throws InvalidStateTransitionSignal
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	@SuppressWarnings("unchecked")
	public HDATA selectLastOneByEventName(HKEY keyInfo, String eventName)
			throws NotFoundSignal, InvalidStateTransitionSignal, DuplicateNameSignal, FrameworkErrorSignal
	{
		try
		{
			String sql = OrmStandardEngineUtil.generateSqlStatement("select", keyInfo);
			sql = OrmStandardEngineUtil.generateExtendedSqlStatement("select", keyInfo, "selectLastOneByEventName", sql);
			Object[] keyArgs = OrmStandardEngineUtil.getPrimaryNoTimeKeyForHistory(keyInfo,2);
			Object[] sqlArgs = new Object[keyArgs.length+1];
			System.arraycopy(keyArgs, 0, sqlArgs, 0, keyArgs.length);
			sqlArgs[sqlArgs.length-1] = eventName;
			List resultList = IDMFrameServiceProxy.getSqlTemplate().queryForList(sql, sqlArgs);
			
			if (resultList.size() == 0){
				throw new NotFoundSignal(getClass(), "selectLastOneByEventName", ObjectUtil.getString(keyInfo)+ ",eventName="
					+ eventName, sql);
			}
			HDATA result =
					(HDATA) validateResultSet(resultList, ObjectUtil.getString(keyInfo) + ";eventName=" + eventName, "");
			return result;
		} catch (Throwable e)
		{
			throw ExceptionNotify.getNotifyException(e);
		}
	}

	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @param fromTime
	 *            java.sql.Timestamp
	 * @param toTime
	 *            java.sql.Timestamp
	 * @return List<DATA> List<extends com.cim.idm.framework.orm.info.DataInfo>
	 * @throws NotFoundSignal
	 * @throws InvalidStateTransitionSignal
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	@SuppressWarnings("unchecked")
	public List<HDATA> selectBetweenTime(HKEY keyInfo, Timestamp fromTime, Timestamp toTime)
			throws NotFoundSignal, InvalidStateTransitionSignal, DuplicateNameSignal, FrameworkErrorSignal
	{
		try
		{
			String sql = OrmStandardEngineUtil.generateSqlStatement("select", keyInfo);
			sql = OrmStandardEngineUtil.generateExtendedSqlStatement("select", keyInfo, "selectBetweenTime", sql);
			
			Object[] bindSet = OrmStandardEngineUtil.getPrimaryNoTimeKeyForHistory(keyInfo,1);			
			bindSet = ObjectUtil.appendObjects(bindSet, fromTime);
			bindSet = ObjectUtil.appendObjects(bindSet, toTime);
			List resultList = IDMFrameServiceProxy.getSqlTemplate().queryForList(sql, bindSet);
			if (resultList.size() == 0)
			{
				String keyString =
						ObjectUtil.getString(keyInfo)
							+ ",fromTime="
							+ TimeUtils.getEventTimeKeyFromTimestamp(fromTime)
							+ ",toTime="
							+ TimeUtils.getEventTimeKeyFromTimestamp(toTime);
				throw new NotFoundSignal(getClass(), "selectBetweenTime", keyString, sql);
			}
			List<HDATA> result = (List<HDATA>) validateResultSet(resultList, ObjectUtil.getString(keyInfo), "arraylist");
			return result;
		} catch (Throwable e)
		{
			throw ExceptionNotify.getNotifyException(e);
		}
	}

	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @param fromTime
	 *            java.sql.Timestamp
	 * @param toTime
	 *            java.sql.Timestamp
	 * @param eventName
	 *            String
	 * @return List<DATA> List<extends com.cim.idm.framework.orm.info.DataInfo>
	 * @throws NotFoundSignal
	 * @throws InvalidStateTransitionSignal
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	@SuppressWarnings("unchecked")
	public List<HDATA> selectBetweenTimeByEventName(HKEY keyInfo, Timestamp fromTime, Timestamp toTime, String eventName)
			throws NotFoundSignal, InvalidStateTransitionSignal, DuplicateNameSignal, FrameworkErrorSignal
	{
		try
		{
			String sql = OrmStandardEngineUtil.generateSqlStatement("select", keyInfo);
			sql = OrmStandardEngineUtil.generateExtendedSqlStatement("select", keyInfo,	"selectBetweenTimeByEventName", sql);
			Object[] bindSet = OrmStandardEngineUtil.getPrimaryNoTimeKeyForHistory(keyInfo,1);
			bindSet = ObjectUtil.appendObjects(bindSet, fromTime);
			bindSet = ObjectUtil.appendObjects(bindSet, toTime);
			bindSet = ObjectUtil.appendObjects(bindSet, eventName);
			List resultList = IDMFrameServiceProxy.getSqlTemplate().queryForList(sql, bindSet);
			if (resultList.size() == 0)
			{
				String keyString =
						ObjectUtil.getString(keyInfo)
							+ ",fromTime="
							+ TimeUtils.getEventTimeKeyFromTimestamp(fromTime)
							+ ",toTime="
							+ TimeUtils.getEventTimeKeyFromTimestamp(toTime)
							+ ",eventName="
							+ eventName;
				throw new NotFoundSignal(getClass(), "selectBetweenTimeByEventName", keyString, sql);
			}
			List<HDATA> result = (List<HDATA>) validateResultSet(resultList, ObjectUtil.getString(keyInfo), "arraylist");
			return result;
		} catch (Throwable e)
		{
			throw ExceptionNotify.getNotifyException(e);
		}
	}

	/**
	 * 
	 * @param keyInfo
	 *            extends com.cim.idm.framework.orm.info.KeyInfo
	 * @return int
	 * @throws NotFoundSignal
	 * @throws InvalidStateTransitionSignal
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	public int deleteLastOne(HKEY keyInfo)
			throws NotFoundSignal, InvalidStateTransitionSignal, DuplicateNameSignal, FrameworkErrorSignal
	{
		IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
		try
		{
			String sql = OrmStandardEngineUtil.generateSqlStatement("delete", keyInfo);
			sql = OrmStandardEngineUtil.generateExtendedSqlStatement("delete", keyInfo, "deleteLastOne", sql);
			int result = IDMFrameServiceProxy.getSqlTemplate().update(sql, OrmStandardEngineUtil.getPrimaryNoTimeKeyForHistory(keyInfo,2));
			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();
			return result;
		} catch (Throwable e)
		{
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			throw ExceptionNotify.getNotifyException(e);
		}
	}

	/**
	 * 
	 * @param dataInfo
	 *            extends com.cim.idm.framework.orm.info.DataInfo
	 * @return int
	 * @throws DuplicateNameSignal
	 * @throws FrameworkErrorSignal
	 * @test
	 * @code
	 * @endcode
	 */
	public int insert(HDATA dataInfo) throws DuplicateNameSignal, FrameworkErrorSignal
	{
//		String cancelFlag = (String) ObjectUtil.getFieldValue(dataInfo, "cancelFlag");
//		if (StringUtils.isEmpty(cancelFlag))
//			ObjectUtil.setFieldValue(dataInfo, "cancelFlag", "A");
//
//		ObjectUtil.setFieldValue(dataInfo, "cancelTimeKey", "");

		return executeInsert(dataInfo, null);
	}

	private int executeInsert(HDATA dataInfo, Timestamp currentTimeStamp)
	{
		IDMFrameServiceProxy.getTxDataSourceManager().beginTransaction();
		try
		{
			if (currentTimeStamp == null)
				currentTimeStamp = TimeUtils.getCurrentTimestamp();

			Object eventTime = ObjectUtil.getFieldValue(dataInfo, "eventTime");
			if (eventTime == null || eventTime.equals(new Timestamp(0)))
				ObjectUtil.setFieldValue(dataInfo, "eventTime", currentTimeStamp);

			String timeKey = String.valueOf(ObjectUtil.getFieldValue(dataInfo.getKey(), "timeKey"));
			if (StringUtils.isEmpty(timeKey))
				ObjectUtil.setFieldValue(dataInfo.getKey(), "timeKey",
					TimeUtils.getEventTimeKeyFromTimestamp(currentTimeStamp));

//			Object systemTime = ObjectUtil.getFieldValue(dataInfo, "systemTime");
//			if (systemTime == null || systemTime.equals(new Timestamp(0)))
//				ObjectUtil.setFieldValue(dataInfo, "systemTime", currentTimeStamp);

			int result = (Integer) this.execute("insert", dataInfo);

			IDMFrameServiceProxy.getTxDataSourceManager().commitTransaction();

			return result;
		} catch (Throwable e)
		{
			IDMFrameServiceProxy.getTxDataSourceManager().rollbackTransaction();
			throw ExceptionNotify.getNotifyException(e);
		}

	}
}
