package com.cim.idm.framework.transaction.impl;

import java.util.Stack;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.DefaultTransactionDefinition;

import com.cim.idm.framework.transaction.IsolationLevel;
import com.cim.idm.framework.transaction.IDMFrameTransactionManager;
import com.cim.idm.framework.transaction.PropagationBehavior;



public class IDMFrameTransactionManagerImpl implements IDMFrameTransactionManager
{
	private static Log						log					= LogFactory.getLog( IDMFrameTransactionManager.class );

	private boolean							autoManaged			= false;

	private PlatformTransactionManager		transactionManager;
	private PropagationBehavior				propagationBehavior	= PropagationBehavior.PROPAGATION_REQUIRED;
	private IsolationLevel					isolationLevel		= IsolationLevel.ISOLATION_READ_COMMITTED;
	private int								transactionTimeout	= 30;

	private ThreadLocal<TransactionInfo>	transactionInfo		= new ThreadLocal<TransactionInfo>();

	/**
	 * 
	 * @param propagationBehavior  PROPAGATION_REQUIRED |  PROPAGATION_NESTED
	 */
	public void setPropagationBehavior( String propagationBehavior )
	{
		if(!StringUtils.isEmpty(propagationBehavior))
		this.propagationBehavior = PropagationBehavior.valueOf( propagationBehavior );
	}
	


	@Override
	public String getPropagationBehavior()
	{
		return propagationBehavior.toString();
	}
	
	public void setIsolationLevel( String isolationLevel )
	{
		if(!StringUtils.isEmpty(isolationLevel))
		this.isolationLevel = IsolationLevel.valueOf( isolationLevel );
	}
	@Override
	public String getIsolationLevel()
	{
		return isolationLevel.toString();
	}
	/**
	 * 
	 * @param transactionTimeout Unit: Seconds
	 */
	public void setTransactionTimeout( int transactionTimeout )
	{
		this.transactionTimeout = transactionTimeout;
	}
	@Override
	public int getTransactionTimeout()
	{
		return transactionTimeout;
	}
	
	public void setAutoManaged( boolean autoManaged )
	{
		this.autoManaged = autoManaged;
	}
	@Override
	public boolean isAutoManaged()
	{
		return autoManaged;
	}
	
	public void setTransactionManager( PlatformTransactionManager transactionManager )
	{
		this.transactionManager = transactionManager;
	}

	@Override
	public void beginTransaction()
	{
		beginTransaction( this.transactionTimeout );
	}

	@Override
	public void beginTransaction( int timeout )
	{
		beginTransaction( this.propagationBehavior, this.isolationLevel, timeout );
	}

	@Override
	public void beginTransaction( PropagationBehavior propagation )
	{
		beginTransaction( propagation, this.transactionTimeout );
	}

	@Override
	public void beginTransaction( PropagationBehavior propagation, int timeout )
	{
		beginTransaction( propagation, this.isolationLevel, timeout );
	}

	@Override
	public void beginTransaction( IsolationLevel isolationLevel )
	{
		beginTransaction( isolationLevel, this.transactionTimeout );
	}

	@Override
	public void beginTransaction( IsolationLevel isolationLevel, int timeout )
	{
		beginTransaction( this.propagationBehavior, isolationLevel, timeout );
	}

	@Override
	public void beginTransaction( PropagationBehavior propagation, IsolationLevel isolationLevel )
	{
		beginTransaction( propagation, isolationLevel, this.transactionTimeout );
	}

	@Override
	public void beginTransaction(PropagationBehavior propagation,
			IsolationLevel isolationLevel, int timeout) {
		TransactionInfo trxInfo = getTransactionInfo();

		DefaultTransactionDefinition transactiondef = new DefaultTransactionDefinition();
		transactiondef.setPropagationBehavior(propagation.value());
		transactiondef.setIsolationLevel(isolationLevel.value());
		transactiondef.setTimeout(timeout);

		Stack<TransactionStatus> statusStack = trxInfo.getStatusStack();

		TransactionStatus transactionStatus = null;
		try {
			transactionStatus = this.transactionManager.getTransaction(transactiondef);
		} catch (TransactionException txException) {
			log.error(
					"TransactionException: in case of lookup, creation, or system errors\n"
							+ "OR IllegalTransactionStateException: if the given transaction definition "
							+ "cannot be executed (for example, if a currently active transaction is "
							+ "in conflict with the specified propagation behavior)",
					txException);
			// throw this Exception to interrupt this function, this is a RunTimeException
			throw txException;
		}
		if(transactionStatus==null){
			log.error("Can not get a currently active transaction or create a new one, according to the "+propagation.value()+" propagation behavior. or in case of lookup, creation, or system errors");
			throw new RuntimeException("Can not get a currently active transaction or create a new one, according to the "+propagation.value()+" propagation behavior. or in case of lookup, creation, or system errors");
		}else {
			statusStack.push(transactionStatus);
		}
		trxInfo.getPropagationStack().push(propagation);
		trxInfo.getIsolationStack().push(isolationLevel);

		if (statusStack.size() == 1)
			log.info("Transaction has been started. " + propagation + ", "
					+ isolationLevel + ", " + timeout + " Sec");
		else if (log.isDebugEnabled())
			log.debug("SubTransaction has been started. " + propagation + ", "
					+ isolationLevel);

		logTransactionRef(statusStack.size(), null, null, false);
	}


	@Override
	public void commitTransaction()
	{
		TransactionInfo trxInfo = this.transactionInfo.get();
		if ( !isTransactionValid( trxInfo )) return;

		Stack<TransactionStatus> statusStack = trxInfo.getStatusStack();
		if ( !isTransactionValid( statusStack )) return;

		TransactionStatus currentStatus = statusStack.pop();
		PropagationBehavior propagation = trxInfo.getPropagationStack().pop();
		IsolationLevel isolation = trxInfo.getIsolationStack().pop();

		if ( isNestedCommit( statusStack, propagation, isolation ) ) 
			return;

		if ( isRollbackOnly( statusStack, currentStatus, propagation )) 
			return;

		try
		{
			this.transactionManager.commit( currentStatus );
		}
		catch ( Exception ex )
		{
			log.error( ex, ex );
		}

		logTransactionRef( statusStack.size(), propagation, isolation, true );

		checkFinalAction( statusStack, true, propagation, isolation );
	}


	@Override
	public void commitAllTransactions()
	{
		TransactionInfo trxInfo = this.transactionInfo.get();
		if ( !isTransactionValid( trxInfo )) 
			return;

		Stack<TransactionStatus> statusStack = trxInfo.getStatusStack();
		if ( !isTransactionValid( statusStack )) 
			return;

		int stackSize = statusStack.size();
		for ( int i = 0; i < stackSize; i++ )
		{
			TransactionStatus currentStatus = statusStack.pop();
			PropagationBehavior propagation = trxInfo.getPropagationStack().pop();
			IsolationLevel isolation = trxInfo.getIsolationStack().pop();
			
			if ( isNestedCommit( statusStack, propagation, isolation )) 
				continue;

			if ( isRollbackOnly( statusStack, currentStatus, propagation )) 
				return;

			try
			{
				this.transactionManager.commit( currentStatus );
			}
			catch ( Exception ex )
			{
				log.error( ex, ex );
			}

			logTransactionRef( statusStack.size(), propagation, isolation, true );

			checkFinalAction( statusStack, true, propagation, isolation );
		}
	}


	@Override
	public void rollbackTransaction()
	{
		TransactionInfo trxInfo = this.transactionInfo.get();
		if ( !isTransactionValid( trxInfo )) 
			return;

		Stack<TransactionStatus> statusStack = trxInfo.getStatusStack();
		if ( !isTransactionValid( statusStack )) 
			return;

		doRollback( trxInfo, statusStack );
	}


	@Override
	public void rollbackAllTransactions()
	{
		TransactionInfo trxInfo = this.transactionInfo.get();
		if ( !isTransactionValid( trxInfo )) 
			return;

		Stack<TransactionStatus> statusStack = trxInfo.getStatusStack();
		if ( !isTransactionValid( statusStack )) 
			return;

		int stackSize = statusStack.size();
		for ( int i = 0; i < stackSize; i++ )
			doRollback( trxInfo, statusStack );
	}


	private void doRollback( TransactionInfo trxInfo, Stack<TransactionStatus> statusStack )
	{
		TransactionStatus currentStatus = statusStack.pop();
		PropagationBehavior propagation = trxInfo.getPropagationStack().pop();
		IsolationLevel isolation = trxInfo.getIsolationStack().pop();

		try
		{
			this.transactionManager.rollback( currentStatus );
		}
		catch ( Exception ex )
		{
			log.error( ex, ex );
		}

		logTransactionRef( statusStack.size(), propagation, isolation, false );

		checkFinalAction( statusStack, false, propagation, isolation );
	}


	private TransactionInfo getTransactionInfo()
	{
		TransactionInfo trxInfo = this.transactionInfo.get();

		if ( trxInfo == null )
		{
			trxInfo = new TransactionInfo();
			this.transactionInfo.set( trxInfo );
		}

		return trxInfo;
	}


	private boolean isTransactionValid( TransactionInfo trxInfo )
	{
		if ( trxInfo == null )
		{
			if ( log.isDebugEnabled()) log.debug( "There is no Transaction started." );
			return false;
		}
		else
			return true;
	}


	private boolean isTransactionValid( Stack<TransactionStatus> statusStack )
	{
		if ( statusStack.isEmpty())
		{
			if ( log.isDebugEnabled()) 
				log.debug( "Transaction has already been ended." );
			
			this.transactionInfo.remove();
			return false;
		}
		else
			return true;
	}


	private void logTransactionRef( int refCount, PropagationBehavior propagation, IsolationLevel isolation, boolean isCommitAction )
	{
		if ( log.isDebugEnabled())
		{
			if ( propagation != null && refCount > 0 )
			{
				if ( isCommitAction )
					log.debug( "SubTransaction has been committed. " + propagation + ", " + isolation );
				else
					log.debug( "SubTransaction has been rolled back. " + propagation + ", " + isolation );
			}

			log.debug( "There is(are) " + refCount + " Transaction reference(s) left that should be committed or rolled back." );
		}
	}

	
	

	private void checkFinalAction( Stack<TransactionStatus> statusStack, boolean isCommitAction, PropagationBehavior propagation, IsolationLevel isolation )
	{
		if ( statusStack.isEmpty() )
		{
			this.transactionInfo.remove();

			if ( isCommitAction )
				log.info( "Transaction has been committed. " + propagation + ", " + isolation );
			else
				log.info( "Transaction has been rolled back. " + propagation + ", " + isolation );
		}
	}


	private boolean isNestedCommit( Stack<TransactionStatus> statusStack, PropagationBehavior propagation, IsolationLevel isolation )
	{
		// PROPAGATION_NESTED if middle Commit Not Allowed
		if ( statusStack.size() > 0 && propagation.equals( PropagationBehavior.PROPAGATION_NESTED ) )
		{
			logTransactionRef( statusStack.size(), propagation, isolation, true );
			return true;
		}
		else
			return false;
	}

	
	private boolean isRollbackOnly( Stack<TransactionStatus> statusStack, TransactionStatus currentStatus, PropagationBehavior propagation )
	{
		
		if ( statusStack.size() == 0 && propagation.equals( PropagationBehavior.PROPAGATION_REQUIRED ) )
		{
			if ( currentStatus.isRollbackOnly() )
			{
				this.transactionManager.rollback( currentStatus );
				log.info( "Transaction has been rolled back because it has been marked as rollback-only" );
				this.transactionInfo.remove();
				
				return true;
			}
		}

		return false;
	}

	private class TransactionInfo
	{
		private Stack<TransactionStatus>	statusStack			= new Stack<TransactionStatus>();
		private Stack<PropagationBehavior>	propagationStack	= new Stack<PropagationBehavior>();
		private Stack<IsolationLevel>		isolationStack		= new Stack<IsolationLevel>();

		

		public Stack<TransactionStatus> getStatusStack()
		{
			return statusStack;
		}

		public Stack<PropagationBehavior> getPropagationStack()
		{
			return propagationStack;
		}

		public Stack<IsolationLevel> getIsolationStack()
		{
			return isolationStack;
		}
	}
}
